/*drvVtr10012.c */

/* Author:   Marty Kraimer */
/* Date:     01NOV2001     */

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution.
*************************************************************************/

#include <vxWorks.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <epicsAssert.h>

#include <vme.h>
#include <vxLib.h>
#include <sysLib.h>
#include <logLib.h>
#include <taskLib.h>
#include <semLib.h>
#include <rebootLib.h>

/*Following needed for block transfer requests*/
#include <sysSymTbl.h>
#include "epicsDynLink.h"
#include <vme.h>
#include <sysDma.h>

#include "ellLib.h"
#include "errlog.h"
#include "devLib.h"
#include "drvSup.h"

#include "drvGtr.h"
#include "drvVtr10012.h"

typedef unsigned int uint32;
typedef unsigned short uint16;

#define NSAM10012_8 0x400
#define STATIC static
/* memory map register offsets */
#define RESET      0x00
#define STATUSR    0x02 /* Should be STATUS but vxWorks already defined STATUS*/
#define CONTROL    0x04
#define INTSTATUS  0x06
#define INTSETUP   0x08
#define CLOCK      0x0A
#define ID         0x0C
#define MULPREPOST 0x0E
#define TRIGGER    0x10
#define ARMR       0x12  /* Should be ARM but vxWorks already defined ARM*/
#define DISARM     0x14
#define RESETIRQ   0x16
#define A32BASE    0x1C
#define HGDR       0x20
#define LGDR       0x22
#define HMLC       0x24
#define LMLC       0x26
#define CPTCCDARM  0x30
#define CPTCC      0x32
#define TCOUNTER   0x34

#define ARRAYSIZE 256*1024 /*sample array size for VTR10012*/
#define ARRAYBYTES ARRAYSIZE*sizeof(uint32)
#define BUFLEN 2048

int vtr10012Debug=0;
int vtr10012UseDma = 0;

typedef enum {vtrType10012,vtrType10012_8,vtrType8014} vtrType;
#define vtrNTypes vtrType8014 + 1
static const char *vtrname[vtrNTypes] = {
    "VTR10012","VTR10012_8","VTR8014"
};
static int16 dataMask[vtrNTypes] = {0x0fff,0x0fff,0x3fff};

#define nclockChoices10012 14
static char *clockChoices10012[nclockChoices10012] = {
    "100 MHz","50 MHz","25 MHz","10 MHz",
    "5 MHz","2.5 MHz","1 MHz",
    "Ext","Ext/2","Ext/4","Ext/10",
    "Ext/20","Ext/40","Ext/100"
};
static uint16 clockValue10012[nclockChoices10012] = {
    0x0000, 0x0001, 0x0002, 0x0003,
    0x0004, 0x0005, 0x0006, 
    0x0008, 0x0009, 0x000a, 0x000b,
    0x000c, 0x000d, 0x000e
};

#define nclockChoices8014 4
static char *clockChoices8014[nclockChoices8014] = {
    "80 MHz","40 MHz",
    "Ext","Ext/2"
};
static uint16 clockValue8014[nclockChoices8014] = {
    0x0000, 0x0001,
    0x0008, 0x0009
};

static int nclockChoices[vtrNTypes] = {
    nclockChoices10012,nclockChoices10012,nclockChoices8014
};
static char **clockChoices[vtrNTypes] = {
    clockChoices10012, clockChoices10012, clockChoices8014
};
static uint16 *clockValue[vtrNTypes] = {
    clockValue10012, clockValue10012, clockValue8014
};

typedef enum {triggerSoft,triggerExt,triggerGate} triggerType;
#define ntriggerChoices 3
static char *triggerChoices[ntriggerChoices] =
{
    "soft","extTrigger","extGate"
};
static uint16 triggerMask[ntriggerChoices] = {
    0x0001, 0x0002, 0x0020
};

#define nmultiEventChoices 5
static char *multiEventChoices[nmultiEventChoices] = {
    "1","2","4","8","16"
};

static int numberEvents[nmultiEventChoices] = {1,2,4,8,16};

typedef enum { armDisarm, armPostTrigger, armPrePostTrigger } armType;
#define narmChoices 3
static char *armChoices[narmChoices] = {
    "disarm","postTrigger","prePostTrigger"
};

typedef struct vtrInfo {
    ELLNODE node;
    struct dma *pdma;
    int     card;
    vtrType type;
    char    *a16;
    int     memoffset;
    char    *memory;
    uint32  *buffer;
    int     intVec;
    int     intLev;
    int     indMultiEventNumber;
    armType arm;
    triggerType trigger;
    int     numberPTS;
    int     numberPPS;
    int     numberPTE;
    gtrhandler usrIH;
    void    *handlerPvt;
    void    *userPvt;
    int     numberEvents;
} vtrInfo;

static ELLLIST vtrList;
static int vtrIsInited = 0;
static int isRebooting;
#define isArmed(pvtrInfo) ((readRegister((pvtrInfo),STATUSR)&0x01) ? 1 : 0)

/* Code to support block transfer vme reads */
/*sysDma not available on all platforms. Needed by vtr10012_8 */
typedef struct dma {
    DMA_ID id;
    SEM_ID dmaDone;
    sysDmaStatus psysDmaStatus;
    sysDmaFromVme psysDmaFromVme;
}dma;

static void dmaCallback(void *context)
{
    dma *pdma = (dma *)context;
    semGive(pdma->dmaDone);
}

static dma *dmaFind(void)
{
    SYM_TYPE symType;	
    sysDmaCreate psysDmaCreate = 0;
    sysDmaStatus psysDmaStatus = 0;
    sysDmaFromVme psysDmaFromVme = 0;
    dma *pdma;

    symFindByNameEPICS(sysSymTbl,"_sysDmaCreate",(void *)&psysDmaCreate,&symType);
    if(!psysDmaCreate) return(0);
    symFindByNameEPICS(sysSymTbl,"_sysDmaStatus",(void *)&psysDmaStatus,&symType);
    if(!psysDmaStatus) return(0);
    symFindByNameEPICS(sysSymTbl,"_sysDmaFromVme",(void *)&psysDmaFromVme,&symType);
    if(!psysDmaFromVme) return(0);
    pdma = calloc(1,sizeof(dma));
    if(!pdma) {
        printf("calloc failed\n");
        return(0);
    }
    pdma->id = (*psysDmaCreate)(dmaCallback,pdma);
    if(!pdma->id) {
        printf("vtr10012: sysDmaCreate failed\n");
        free(pdma);
        pdma = 0;
    }
    pdma->dmaDone = semBCreate(SEM_Q_PRIORITY,SEM_EMPTY);
    pdma->psysDmaStatus = psysDmaStatus;
    pdma->psysDmaFromVme = psysDmaFromVme;
    return(pdma);;
}

static int dmaRead(dma *pdma,uint32 vmeaddr,uint32 *buffer,int len)
{
    STATUS status;

    if(vtr10012Debug)
        printf("dmaRead(%p,%x,%p,%d)\n",pdma->id,vmeaddr,buffer,len);

    status = (*pdma->psysDmaFromVme) (pdma->id,
        (void *)buffer,vmeaddr,VME_AM_EXT_SUP_ASCENDING,len,4);
    if(status) {
        printf("vtr10012: dmaRead error %s\n",strerror(errno));
        return(-1);
    }
    semTake(pdma->dmaDone,WAIT_FOREVER);
    status = (*pdma->psysDmaStatus)(pdma->id);
    if(status) {
        printf("vtr10012: dmaDone error %s\n",strerror(errno));
        return(-1);
    }
    if(vtr10012Debug) {
        printf("dmaRead OK, vmeaddr %8.8x len %d\n",vmeaddr,len);
    }
    return(0);
}

static void writeRegister(vtrInfo *pvtrInfo, int offset,uint16 value)
{
    char *a16 = pvtrInfo->a16;
    uint16 *reg;

    reg = (uint16 *)(a16+offset);
    *reg = value;
}

static uint16 readRegister(vtrInfo *pvtrInfo, int offset)
{
    char *a16 = pvtrInfo->a16;
    uint16 *reg;
    uint16 value;

    reg = (uint16 *)(a16+offset);
    value = *reg;
    return(value);
}

static void writeLocation(vtrInfo *pvtrInfo,int value)
{
    writeRegister(pvtrInfo,HMLC,(value>>16)&0xffff);
    writeRegister(pvtrInfo,LMLC,value&0xffff);
}

static void writeGate(vtrInfo *pvtrInfo,int value)
{
    writeRegister(pvtrInfo,HGDR,(value>>16)&0xffff);
    writeRegister(pvtrInfo,LGDR,value&0xffff);
}

STATIC uint32 readTriggerCounter(vtrInfo *pvtrInfo)
{
    uint16 low,high;

    if(pvtrInfo->type==vtrType10012_8) return(1024);
    high = readRegister(pvtrInfo,TCOUNTER);
    low = readRegister(pvtrInfo,TCOUNTER);
    return(high<<16 | low);
}

static int vtrReboot(int boot_type)
{
    vtrInfo  *pvtrInfo;

    isRebooting = 1;
    pvtrInfo = (vtrInfo *)ellFirst(&vtrList);
    while(pvtrInfo) {
        writeRegister(pvtrInfo,RESET,0x01);
        pvtrInfo = (vtrInfo *)ellNext(&pvtrInfo->node);
    }
    vtrIsInited = 0;
    return(0);
}
    
static void initialize()
{
    if(vtrIsInited) return;
    vtrIsInited=1;
    isRebooting = 0;
    ellInit(&vtrList);
    rebootHookAdd(vtrReboot);
}

void vtr10012IH(vtrInfo *pvtrInfo)
{

    if(isRebooting || (pvtrInfo->arm == armDisarm)) {
        writeRegister(pvtrInfo,DISARM,1); 
        return;
    }
    if(pvtrInfo->type!=vtrType10012_8) {
        uint16 regCPTCC = readRegister(pvtrInfo,CPTCC);
        if(pvtrInfo->arm == armPostTrigger) {
            if(regCPTCC < pvtrInfo->numberPTE) return;
        } else if(pvtrInfo->arm == armPrePostTrigger) {
            if(regCPTCC < pvtrInfo->numberEvents)  return;
        }
    }
    writeRegister(pvtrInfo,DISARM,1); 
    if(pvtrInfo->usrIH) (*pvtrInfo->usrIH)(pvtrInfo->handlerPvt);
}

STATIC void vtrinit(gtrPvt pvt)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
    long status;
    
    status = devConnectInterrupt(intVME,pvtrInfo->intVec,
        vtr10012IH,(void *)pvtrInfo);
    if(status) {
        errMessage(status,"vtrConfig devConnectInterrupt failed\n");
        return;
    }
    status = devEnableInterruptLevel(intVME,pvtrInfo->intLev);
    if(status) {
        errMessage(status,"init devEnableInterruptLevel failed\n");
    }
    writeRegister(pvtrInfo,RESET,1);
    writeRegister(pvtrInfo,INTSTATUS,pvtrInfo->intVec);
    writeRegister(pvtrInfo,INTSETUP,pvtrInfo->intLev);
    writeRegister(pvtrInfo,A32BASE,(pvtrInfo->memoffset)>>24);
    return;
}

STATIC void vtrreport(gtrPvt pvt,int level)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;

    printf("%s card %d a16 %p memory %p intVec %2.2x intLev %d\n",
        vtrname[pvtrInfo->type],pvtrInfo->card,
        pvtrInfo->a16,pvtrInfo->memory,
        pvtrInfo->intVec,pvtrInfo->intLev);
}

STATIC gtrStatus vtrclock(gtrPvt pvt, int value)
{ 
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
    int nchoices;

    if(isArmed(pvtrInfo)) return(gtrStatusBusy);
    if(isRebooting) taskSuspend(0);
    nchoices = nclockChoices[pvtrInfo->type];
    if(value<0 || value>=nchoices) return(gtrStatusError);
    writeRegister(pvtrInfo,CLOCK,clockValue[pvtrInfo->type][value]);
    return(gtrStatusOK);
}

STATIC gtrStatus vtrtrigger(gtrPvt pvt, int value)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
    uint16 reg;

    if(isRebooting) taskSuspend(0);
    if(value<0 || value>ntriggerChoices) return(gtrStatusError);
    if(isArmed(pvtrInfo)) return(gtrStatusBusy);
    pvtrInfo->trigger = value;
    reg = readRegister(pvtrInfo,CONTROL);
    reg = (reg & ~(0x0023)) | triggerMask[pvtrInfo->trigger];
    writeRegister(pvtrInfo,CONTROL,reg);
    return(gtrStatusOK);
}

STATIC gtrStatus vtrmultiEvent(gtrPvt pvt, int value)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;

    if(value<0 || value>=nmultiEventChoices) return(gtrStatusError);
    if(pvtrInfo->type==vtrType10012_8 && value!=0) return(gtrStatusError);
    if(isArmed(pvtrInfo)) return(gtrStatusBusy);
    if(isRebooting) taskSuspend(0);
    pvtrInfo->indMultiEventNumber = value;
    pvtrInfo->numberEvents = numberEvents[pvtrInfo->indMultiEventNumber];
    return(gtrStatusOK);
}

STATIC gtrStatus vtrnumberPTS(gtrPvt pvt, int value)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;

    if(isArmed(pvtrInfo)) return(gtrStatusBusy);
    if(pvtrInfo->type==vtrType10012_8 && value>1024) return(gtrStatusError);
    pvtrInfo->numberPTS = value;
    return(gtrStatusOK);
}

STATIC gtrStatus vtrnumberPPS(gtrPvt pvt, int value)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;

    if(isArmed(pvtrInfo)) return(gtrStatusBusy);
    if(pvtrInfo->type==vtrType10012_8) return(gtrStatusError);
    pvtrInfo->numberPPS = value;
    return(gtrStatusOK);
}

STATIC gtrStatus vtrnumberPTE(gtrPvt pvt, int value)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;

    if(isArmed(pvtrInfo)) return(gtrStatusBusy);
    if(pvtrInfo->type==vtrType10012_8) return(gtrStatusError);
    pvtrInfo->numberPTE = value;
    return(gtrStatusOK);
}

STATIC gtrStatus vtrarm(gtrPvt pvt, int typ)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
    uint16 reg,regControl;
    armType arm = (armType)typ;

    if(pvtrInfo->type==vtrType10012_8
    &&(arm!=armDisarm && arm!=armPostTrigger)) return(gtrStatusError);
    pvtrInfo->arm = arm;
    writeRegister(pvtrInfo,DISARM,1);
    writeRegister(pvtrInfo,RESETIRQ,1);
    if(pvtrInfo->arm==armDisarm) return(gtrStatusOK);
    writeLocation(pvtrInfo,0);
    writeGate(pvtrInfo,pvtrInfo->numberPTS);
    reg = readRegister(pvtrInfo,INTSETUP);
    writeRegister(pvtrInfo,INTSETUP,reg|0x0008); /*IRQ Enable*/
    regControl = ~0x0048 & readRegister(pvtrInfo,CONTROL);
    writeRegister(pvtrInfo,CONTROL,regControl);
    writeRegister(pvtrInfo,CPTCC,1);
    writeRegister(pvtrInfo,TCOUNTER,1);
    switch(arm) {
    case armPostTrigger:
        writeRegister(pvtrInfo,MULPREPOST,0);
        writeRegister(pvtrInfo,CONTROL,regControl);
        writeRegister(pvtrInfo,ARMR,1);
        break;
    case armPrePostTrigger: {
        uint16 multi = 0;
        if(pvtrInfo->indMultiEventNumber>0)
            multi = 0x0004|(pvtrInfo->indMultiEventNumber - 1);
        writeRegister(pvtrInfo,MULPREPOST,multi);
        regControl |= 0x0048;
        writeRegister(pvtrInfo,CONTROL,regControl);
        writeRegister(pvtrInfo,ARMR,1);
        break;
    }
    default:
        printf("vtrarm: Illegal value\n");
    }
    return(gtrStatusOK);
}

STATIC gtrStatus vtrsoftTrigger(gtrPvt pvt)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
    writeRegister(pvtrInfo,TRIGGER,1);
    return(gtrStatusOK);
}

STATIC void readContiguous(vtrInfo *pvtrInfo,
    gtrchannel *phigh,gtrchannel *plow,uint32 *pmemory,
    int nmax,int *nskipHigh, int *nskipLow)
{
    int16 high,low,mask;
    int ind;
    int bufOffset = BUFLEN;

    if(vtr10012Debug)
        printf("readContiguous pmemory %p nmax %d\n",pmemory,nmax);
    mask = dataMask[pvtrInfo->type];
    for(ind=0; ind<nmax; ind++) {
        uint32 word;

        if(pvtrInfo->pdma && vtr10012UseDma) {
            if(bufOffset>=BUFLEN) {
                int status;
                uint32 VMEaddr,bytesRemaining,bytesMax,nbytes;
                VMEaddr = pvtrInfo->memoffset
                    + ((char *)(pmemory) - pvtrInfo->memory)
                    + ind * sizeof(uint32);
                bytesRemaining = (nmax-ind)*sizeof(uint32);
                bytesMax = BUFLEN*sizeof(uint32);
                nbytes = (bytesRemaining<bytesMax) ? bytesRemaining : bytesMax;
                status = dmaRead(pvtrInfo->pdma,VMEaddr,pvtrInfo->buffer,nbytes);
                if(status) break;
                bufOffset = 0;
            }
            word = pvtrInfo->buffer[bufOffset++];
        } else {
            word = pmemory[ind];
        }
        if(*nskipHigh>0) {
            --*nskipHigh;
        } else if(phigh->ndata<phigh->len) {
            high = (word>>16)&mask;
            (phigh->pdata)[phigh->ndata++] = high;
        }
        if(*nskipLow>0) {
            --*nskipLow;
        } else if(plow->ndata<plow->len) {
            low = word&mask;
            (plow->pdata)[plow->ndata++] = low;
        }
        if((phigh->ndata>=phigh->len) && (plow->ndata>=plow->len)) break;
    }
}

STATIC gtrStatus readPostTrigger(vtrInfo *pvtrInfo,gtrchannel **papgtrchannel)
{
    int indgroup;

    for(indgroup=0; indgroup<4; indgroup++) {
        uint32 *pgroup = (uint32 *)(pvtrInfo->memory + indgroup*0x00400000);
        gtrchannel *phigh,*plow;
        int ndata,nskipHigh,nskipLow;

        ndata = pvtrInfo->numberPTS * pvtrInfo->numberPTE;
        if(ndata>ARRAYSIZE) ndata = ARRAYSIZE;
        phigh = papgtrchannel[indgroup + 4];
        plow = papgtrchannel[indgroup];
        nskipHigh = nskipLow = 0;
        readContiguous(pvtrInfo,phigh,plow,pgroup,ndata,&nskipHigh,&nskipLow);
    }
    return(gtrStatusOK);
}

STATIC gtrStatus readPrePostTrigger(vtrInfo *pvtrInfo,gtrchannel **papgtrchannel)
{
    int numberPPS = pvtrInfo->numberPPS;
    int nevents = pvtrInfo->numberEvents;
    int indevent,eventsize;

    if(nevents != readRegister(pvtrInfo,CPTCCDARM)) {
        printf("drvVtr10012: numberEvents %d but CPTCCDARM %d\n",
            nevents,readRegister(pvtrInfo,CPTCCDARM));
    }
    writeRegister(pvtrInfo,TCOUNTER,1);
    eventsize = ARRAYSIZE/nevents;
    if(numberPPS>eventsize) numberPPS = eventsize;
    if(numberPPS==0) numberPPS = eventsize;
    for(indevent=0; indevent<nevents; indevent++) {
        int location = readTriggerCounter(pvtrInfo);
        int indgroup;

        location -= indevent*eventsize;
        if(location<0 || location>=eventsize) {
            printf("location %d but eventsize %d\n",location,eventsize);
            continue;
        }
        /*readTriggerCounter returns location of last value saved*/
        location += 1;
        if(location==eventsize) location = 0;
        for(indgroup=0; indgroup<4; indgroup++) {
            uint32 *pgroup = (uint32 *)(pvtrInfo->memory + indgroup*0x00400000);
            uint32 *pmemory = pgroup + indevent*eventsize;
            gtrchannel *phigh,*plow;
            int nskipHigh,nskipLow,nhigh,nlow,nmax;
    
            phigh = papgtrchannel[indgroup + 4];
            plow = papgtrchannel[indgroup];
            nhigh = phigh->len - phigh->ndata;
            if(nhigh>numberPPS) nhigh = numberPPS;
            nlow = plow->len - plow->ndata;
            if(nlow>numberPPS) nlow = numberPPS;
            nmax = (nhigh>nlow) ? nhigh : nlow;
            if(nmax<=0) continue;
            nskipHigh = nmax - nhigh;
            nskipLow = nmax - nlow;
            if(location < nmax) {
                int nend,nbeg;

                nend = nmax - location;
                nbeg = nmax - nend;
                readContiguous(pvtrInfo,phigh,plow,
                    (pmemory + eventsize - nend),nend,
                    &nskipHigh,&nskipLow);
                readContiguous(pvtrInfo,phigh,plow,
                    pmemory,nbeg,&nskipHigh,&nskipLow);
            } else {
                readContiguous(pvtrInfo,phigh,plow,
                    (pmemory + location - nmax),nmax,
                    &nskipHigh,&nskipLow);
            }
        }
    }
    return(gtrStatusOK);
}

STATIC gtrStatus vtrreadMemory(gtrPvt pvt,gtrchannel **papgtrchannel)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
    int indgroup;

    if(pvtrInfo->arm==armDisarm) return(gtrStatusOK);
    if(pvtrInfo->type==vtrType10012_8 && pvtrInfo->arm!=armPostTrigger)
        return(gtrStatusError);
    for(indgroup=0; indgroup<4; indgroup++) {
        gtrchannel *phigh;
        gtrchannel *plow;

        phigh = papgtrchannel[indgroup + 4];
        plow = papgtrchannel[indgroup];
        phigh->ndata=0;
        plow->ndata=0;
    }
    if(pvtrInfo->arm==armPostTrigger) {
        readPostTrigger(pvtrInfo,papgtrchannel);
    } else if(pvtrInfo->arm==armPrePostTrigger) {
        readPrePostTrigger(pvtrInfo,papgtrchannel);
    }  else { printf("Illegal arm request\n"); }
    return(gtrStatusOK);
}

STATIC gtrStatus vtrgetLimits(gtrPvt pvt,int16 *rawLow,int16 *rawHigh)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
    *rawLow = 0;
    *rawHigh = dataMask[pvtrInfo->type] + 1;
    return(gtrStatusOK);
}

STATIC gtrStatus vtrregisterHandler(gtrPvt pvt,
     gtrhandler usrIH,void *handlerPvt)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
   
    pvtrInfo->usrIH = usrIH;
    pvtrInfo->handlerPvt = handlerPvt;
    return(gtrStatusOK);
}

STATIC int vtrnumberChannels(gtrPvt pvt)
{
    return(8);
}

STATIC gtrStatus vtrclockChoices(gtrPvt pvt,int *number,char ***choice)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;

    *number = nclockChoices[pvtrInfo->type];
    *choice = clockChoices[pvtrInfo->type];
    return(gtrStatusOK);
}

STATIC gtrStatus vtrarmChoices(gtrPvt pvt,int *number,char ***choice)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
   
    *number = (pvtrInfo->type==vtrType10012_8) ? 2 : narmChoices;
    *choice = armChoices;
    return(gtrStatusOK);
}

STATIC gtrStatus vtrtriggerChoices(gtrPvt pvt,int *number,char ***choice)
{
    *number = ntriggerChoices;
    *choice = triggerChoices;
    return(gtrStatusOK);
}

STATIC gtrStatus vtrmultiEventChoices(gtrPvt pvt,int *number,char ***choice)
{
    vtrInfo *pvtrInfo = (vtrInfo *)pvt;
   
    *number = (pvtrInfo->type==vtrType10012_8) ? 1 : nmultiEventChoices;
    *choice = multiEventChoices;
    return(gtrStatusOK);
}

static gtrops vtr10012ops = {
vtrinit, 
vtrreport, 
vtrclock,
vtrtrigger, 
vtrmultiEvent, 
vtrnumberPTS,
vtrnumberPPS,
vtrnumberPTE,
vtrarm,
vtrsoftTrigger,
vtrreadMemory,
vtrgetLimits,
vtrregisterHandler,
vtrnumberChannels,
vtrclockChoices,
vtrarmChoices,
vtrtriggerChoices,
vtrmultiEventChoices,
0,0,0,0,0
};


int vtr10012Config(int card,
    int a16offset,unsigned int memoffset,
    int intVec,int intLev)
{
    char *a16;
    gtrops *pgtrops;
    uint16 probeValue = 0;
    vtrInfo *pvtrInfo;
    long status;
    vtrType type;
    char *memory;
    dma *pdma = 0;

    if(!vtrIsInited) initialize();
    if(gtrFind(card,&pgtrops)) {
        printf("card is already configured\n"); return(0);
    }
    if((a16offset&0xff00)!=a16offset) {
        printf("vtrConfig: illegal a16offset. Must be multiple of 0x0100\n");
        return(0);
    }
    status = devRegisterAddress("vtr",atVMEA16,
        (void *)a16offset,0x100,(void **)&a16);
    if(status) {
        errMessage(status,"vtrConfig devRegisterAddress failed\n");
        return(0);
    }
    if(vxMemProbe(a16+ID,VX_READ,2,(char *)&probeValue)!=OK) {
        printf("vtrConfig: no card at %p\n",a16);
        return(0);
    }
    pdma = dmaFind();
    probeValue = (probeValue>>10);
    if(probeValue==7) {
        type = vtrType10012;
    } else if(probeValue==8) {
        type = vtrType10012_8;
        if(!pdma) {
            printf("dma is not available but vtr10012_8 requires it.\n");
            return(0);
        }
    } else if(probeValue==9) {
        type = vtrType8014;
    } else {
        printf("ID %x not vtrType10012 or vtrType10012_8 or vtrType8014\n",probeValue);
        return(0);
    }
    if((memoffset&0xFF000000)!=memoffset) {
        printf("memoffset must be multiple of 0xFF000000\n");
        return(0);
    }
    status = devRegisterAddress(vtrname[type],atVMEA32,
        (void *)memoffset,0x01000000,(void **)&memory);
    if(status) {
        errMessage(status,"vtrConfig devRegisterAddress failed for memory\n");
        return(0);
    }
    pvtrInfo = calloc(1,sizeof(vtrInfo));
    if(!pvtrInfo) {
        printf("vtrConfig: calloc failed\n");
        return(0);
    }
    if(pdma) {
        pvtrInfo->buffer = calloc(BUFLEN,sizeof(uint32));
        if(!pvtrInfo->buffer) {
            printf("vtrConfig: calloc failed\n");
            return(0);
        }
    }
    pvtrInfo->pdma = pdma;
    pvtrInfo->card = card;
    pvtrInfo->type = type;
    pvtrInfo->a16 = a16;
    pvtrInfo->memoffset = memoffset;
    pvtrInfo->memory = memory;
    pvtrInfo->intVec = intVec;
    pvtrInfo->intLev = intLev;
    pvtrInfo->numberPTE = 1;
    if(type==vtrType10012_8) {
        pvtrInfo->numberPTS = NSAM10012_8;
        pvtrInfo->numberPTE = 1;
    }
    pvtrInfo->numberEvents = 1;
    ellAdd(&vtrList,&pvtrInfo->node);
    gtrRegisterDriver(card,vtrname[type],&vtr10012ops,pvtrInfo);
    return(0);
}
